@@ -36,6 +36,11 @@
 
 #include <girepository.h>
 
+static JSBool gjs_value_from_g_value_internal(JSContext    *context,
+                                              jsval        *value_p,
+                                              const GValue *gvalue,
+                                              gboolean      no_copy);
+
 static void
 closure_marshal(GClosure        *closure,
                 GValue          *return_value,
@@ -49,6 +54,7 @@ closure_marshal(GClosure        *closure,
     jsval *argv;
     jsval rval;
     int i;
+    GSignalQuery signal_query = { 0, };
 
     gjs_debug_marshal(GJS_DEBUG_GCLOSURE,
                       "Marshal closure %p",
@@ -68,9 +74,38 @@ closure_marshal(GClosure        *closure,
     gjs_root_value_locations(context, argv, argc);
     JS_AddRoot(context, &rval);
 
+    if (marshal_data) {
+        /* we are used for a signal handler */
+        guint signal_id;
+
+        signal_id = GPOINTER_TO_UINT(marshal_data);
+
+        g_signal_query(signal_id, &signal_query);
+
+        if (!signal_query.signal_id) {
+            gjs_debug(GJS_DEBUG_GCLOSURE,
+                      "Signal handler being called on invalid signal");
+            goto cleanup;
+        }
+
+        if (signal_query.n_params + 1 != n_param_values) {
+            gjs_debug(GJS_DEBUG_GCLOSURE,
+                      "Signal handler being called with wrong number of parameters");
+            goto cleanup;
+        }
+    }
+
     for (i = 0; i < argc; ++i) {
         const GValue *gval = &param_values[i];
-        if (!gjs_value_from_g_value(context, &argv[i], gval)) {
+        gboolean no_copy;
+
+        no_copy = FALSE;
+
+        if (i >= 1 && signal_query.signal_id) {
+            no_copy = (signal_query.param_types[i - 1] & G_SIGNAL_TYPE_STATIC_SCOPE) != 0;
+        }
+
+        if (!gjs_value_from_g_value_internal(context, &argv[i], gval, no_copy)) {
             gjs_debug(GJS_DEBUG_GCLOSURE,
                       "Unable to convert arg %d in order to invoke closure",
                       i);
@@ -100,6 +135,21 @@ closure_marshal(GClosure        *closure,
     JS_RemoveRoot(context, &rval);
 }
 
+GClosure*
+gjs_closure_new_for_signal(JSContext  *context,
+                           JSObject   *callable,
+                           const char *description,
+                           guint       signal_id)
+{
+    GClosure *closure;
+
+    closure = gjs_closure_new(context, callable, description);
+
+    g_closure_set_meta_marshal(closure, GUINT_TO_POINTER(signal_id), closure_marshal);
+
+    return closure;
+}
+
 GClosure*
 gjs_closure_new_marshaled (JSContext    *context,
                            JSObject     *callable,
@@ -139,10 +189,11 @@ gjs_value_guess_g_type(jsval value)
     return G_TYPE_INVALID;
 }
 
-JSBool
-gjs_value_to_g_value(JSContext    *context,
-                     jsval         value,
-                     GValue       *gvalue)
+static JSBool
+gjs_value_to_g_value_internal(JSContext    *context,
+                              jsval         value,
+                              GValue       *gvalue,
+                              gboolean      no_copy)
 {
     GType gtype;
 
@@ -292,7 +343,7 @@ gjs_value_to_g_value(JSContext    *context,
             guint32 length;
 
             if (!gjs_object_require_property(context,
-                                             JSVAL_TO_OBJECT(value),
+                                             JSVAL_TO_OBJECT(value), NULL,
                                              "length",
                                              &length_value) ||
                 !JS_ValueToECMAUint32(context, length_value, &length)) {
@@ -301,12 +352,15 @@ gjs_value_to_g_value(JSContext    *context,
                           gjs_get_type_name(value));
                 return JS_FALSE;
             } else {
+                void *result;
                 char **strv;
 
                 if (!gjs_array_to_strv (context,
                                         value,
-                                        length, (void**)&strv))
+                                        length, &result))
                     return JS_FALSE;
+                /* cast to strv in a separate step to avoid type-punning */
+                strv = result;
                 g_value_take_boxed (gvalue, strv);
             }
         } else {
@@ -324,7 +378,7 @@ gjs_value_to_g_value(JSContext    *context,
         } else if (JSVAL_IS_OBJECT(value)) {
             JSObject *obj;
             obj = JSVAL_TO_OBJECT(value);
-            gboxed = gjs_g_boxed_from_boxed(context, obj);
+            gboxed = gjs_c_struct_from_boxed(context, obj);
         } else {
             gjs_throw(context,
                       "Wrong type %s; boxed type %s expected",
@@ -333,7 +387,10 @@ gjs_value_to_g_value(JSContext    *context,
             return JS_FALSE;
         }
 
-        g_value_set_boxed(gvalue, gboxed);
+        if (no_copy)
+            g_value_set_static_boxed(gvalue, gboxed);
+        else
+            g_value_set_boxed(gvalue, gboxed);
     } else if (g_type_is_a(gtype, G_TYPE_ENUM)) {
         if (JSVAL_IS_INT(value)) {
             GEnumValue *v;
@@ -357,9 +414,7 @@ gjs_value_to_g_value(JSContext    *context,
         }
     } else if (g_type_is_a(gtype, G_TYPE_FLAGS)) {
         if (JSVAL_IS_INT(value)) {
-            if (!_gjs_flags_value_is_valid(context,
-                                           G_FLAGS_CLASS(g_type_class_peek(gtype)),
-                                           JSVAL_TO_INT(value)))
+            if (!_gjs_flags_value_is_valid(context, gtype, JSVAL_TO_INT(value)))
                 return JS_FALSE;
 
             g_value_set_flags(gvalue, value);
@@ -432,9 +487,26 @@ gjs_value_to_g_value(JSContext    *context,
 }
 
 JSBool
-gjs_value_from_g_value(JSContext    *context,
-                       jsval        *value_p,
-                       const GValue *gvalue)
+gjs_value_to_g_value(JSContext    *context,
+                     jsval         value,
+                     GValue       *gvalue)
+{
+    return gjs_value_to_g_value_internal(context, value, gvalue, FALSE);
+}
+
+JSBool
+gjs_value_to_g_value_no_copy(JSContext    *context,
+                             jsval         value,
+                             GValue       *gvalue)
+{
+    return gjs_value_to_g_value_internal(context, value, gvalue, TRUE);
+}
+
+static JSBool
+gjs_value_from_g_value_internal(JSContext    *context,
+                                jsval        *value_p,
+                                const GValue *gvalue,
+                                gboolean      no_copy)
 {
     GType gtype;
 
@@ -492,21 +564,41 @@ gjs_value_from_g_value(JSContext    *context,
         obj = gjs_object_from_g_object(context, gobj);
         *value_p = OBJECT_TO_JSVAL(obj);
     } else if (g_type_is_a(gtype, G_TYPE_BOXED)) {
+        GjsBoxedCreationFlags boxed_flags;
         GIBaseInfo *info;
         void *gboxed;
         JSObject *obj;
 
         gboxed = g_value_get_boxed(gvalue);
+        boxed_flags = GJS_BOXED_CREATION_NONE;
 
         /* The only way to differentiate unions and structs is from
          * their g-i info as both GBoxed */
         info = g_irepository_find_by_gtype(g_irepository_get_default(),
                                            gtype);
+        if (info == NULL) {
+            gjs_throw(context,
+                      "No introspection information found for %s",
+                      g_type_name(gtype));
+            return JS_FALSE;
+        }
 
-        if (g_base_info_get_type(info) == GI_INFO_TYPE_UNION) {
-            obj = gjs_union_from_g_boxed(context, gtype, gboxed);
-        } else {
-            obj = gjs_boxed_from_g_boxed(context, gtype, gboxed);
+        switch (g_base_info_get_type(info)) {
+        case GI_INFO_TYPE_BOXED:
+        case GI_INFO_TYPE_STRUCT:
+            if (no_copy)
+                boxed_flags |= GJS_BOXED_CREATION_NO_COPY;
+            obj = gjs_boxed_from_c_struct(context, (GIStructInfo *)info, gboxed, boxed_flags);
+            break;
+        case GI_INFO_TYPE_UNION:
+            obj = gjs_union_from_c_union(context, (GIUnionInfo *)info, gboxed);
+            break;
+        default:
+            gjs_throw(context,
+                      "Unexpected introspection type %d for %s",
+                      g_base_info_get_type(info),
+                      g_type_name(gtype));
+            return JS_FALSE;
         }
         *value_p = OBJECT_TO_JSVAL(obj);
     } else if (g_type_is_a(gtype, G_TYPE_ENUM)) {
@@ -556,3 +648,11 @@ gjs_value_from_g_value(JSContext    *context,
 
     return JS_TRUE;
 }
+
+JSBool
+gjs_value_from_g_value(JSContext    *context,
+                       jsval        *value_p,
+                       const GValue *gvalue)
+{
+    return gjs_value_from_g_value_internal(context, value_p, gvalue, FALSE);
+}
